/**
  *  \file client/chartmode.h
  *  \brief Starchart Mode Handling
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_CLIENT_CHARTMODE_H
#define PCC_V2_CLIENT_CHARTMODE_H

#include "cpluslib/object.h"
#include "gfx/gfxutil.h"

class WChartModeList;
class GChartLocation;
class GfxCanvas;
struct UIEvent;

enum {
    /** A temporary message. */
    ChartMode_Message     = 0,

    /** Prefix argument. */
    ChartMode_Prefix      = 1,

    /** A primary mode, which changes the keymap completely.
        Example: marker manipulation modes */
    ChartMode_PrimaryMode = 10,

    /** A temporary sub-mode.
        Example: temporary display filters. */
    ChartMode_SubMode     = 20,

    /** A base mode, providing movement etc. */
    ChartMode_BaseMode    = 30
};

/** Starchart mode. This represents a starchart overlay and a set of
    key assignments.

    A starchart can simultaneously be in a multitude of modes,
    represented as a list of modes (WChartModeList). Each mode has a
    kind, only one mode of a kind can be active at a time (this is
    used to avoid uncontrolled mode stacking). */
class WChartMode : public Object {
    friend class WChartModeList;
    WChartModeList* parent;
    WChartMode* next;
    int kind;
 public:
    WChartMode(int kind);
    ~WChartMode();

    WChartMode* getNextMode() const;
    WChartModeList* getParent();

    /** Handle movement. Called on every iteration through the main event loop.
        \param loc Location we're working on. */
    virtual void onMove(GChartLocation& loc) = 0;

    /** Draw below chart. Called before the chart has been drawn.
        \param can Canvas to draw on.
        \param loc Location we're working on. */
    virtual void drawBelow(GfxCanvas& can, GChartLocation& loc) = 0;

    /** Draw cursor. Called first during chart redraw, after the chart has been
        drawn. Should draw the cursor, if the mode has one. Unless this mode
        "hides" the following mode's cursor, it should call the next's drawCursor() as well.
        \param can Canvas to draw on.
        \param loc Location we're working on. */
    virtual void drawCursor(GfxCanvas& can, GChartLocation& loc) = 0;

    /** Draw overlays. Called second during chart redraw. Should draw overlays
        like popups and prompts. Unless this mode "hides" the following mode's
        overlays, it should call the next's drawOverlays() as well.
        \param can Canvas to draw on
        \param loc Location we're working on */
    virtual void drawOverlays(GfxCanvas& can, GChartLocation& loc) = 0;

    /** Handle event. If this mode does not handle the event, it should pass it
        to the next mode (unless this mode explicitly intents to filter the event).
        \param even Event to handle
        \param loc Location we're working on
        \return true iff event was handled, and main loop should not handle it */
    virtual bool handleEvent(const UIEvent& event, GChartLocation& loc) = 0;
};

/** Simple starchart mode. Use this as a base class if you override
    only a few methods of WChartMode. */
class WSimpleChartMode : public WChartMode {
 public:
    WSimpleChartMode(int kind);
    ~WSimpleChartMode();

    void onMove(GChartLocation& loc);
    void drawBelow(GfxCanvas& can, GChartLocation& loc);
    void drawCursor(GfxCanvas& can, GChartLocation& loc);
    void drawOverlays(GfxCanvas& can, GChartLocation& loc);
    bool handleEvent(const UIEvent& event, GChartLocation& loc);
};

/** Starchart mode list. Represents a set of modes the starchart can
    be in. This class offers handling of a mode list, as well as
    tracking of dirtiness status of map objects. */
class WChartModeList {
    WChartMode* current_mode;
    bool changed;

    GfxRect dirty;
 public:
    WChartModeList();
    ~WChartModeList();

    void deleteAllModes();
    void pushMode(WChartMode& which);
    void removeMode(WChartMode& which);
    WChartMode* getModeByKind(int k) const;
    WChartMode* getCurrentMode() const;

    void onMove(GChartLocation& loc);
    void drawBelow(GfxCanvas& can, GChartLocation& loc);
    void drawCursor(GfxCanvas& can, GChartLocation& loc);
    void drawOverlays(GfxCanvas& can, GChartLocation& loc);
    bool handleEvent(const UIEvent& event, GChartLocation& loc);

    void markChanged();
    void clearChanged();
    bool isChanged() const;

    void markDirty(GfxRect r);
    void markClean();
    bool isDirty() const;
    const GfxRect& getDirtyRect() const;
};

#endif
