/**
  *  \file client/lru.h
  *  \brief Least-Recently-Used (and Predefined) Lists
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_CLIENT_LRU_H
#define PCC_V2_CLIENT_LRU_H

#include "cpluslib/string.h"
#include "cpluslib/ptrvector.h"
#include "cpluslib/types.h"

/** Item of an LRU list. Each item has a name and value. The
    interpretation of the value is up to the actual application.

    As a special convention, values may start with a value in brackets
    "[...]", which is ignored when comparing items. */
class LRUListItem {
 public:
    LRUListItem(string_t name, string_t value);
    const string_t& getName() const;
    const string_t& getValue() const;

 private:
    string_t name;
    string_t value;
};

/** LRU (or predefined) list. Contains a list of LRUListItem's, and
    operations to work with them. */
class LRUList {
 public:
    LRUList();
    ~LRUList();

    // array access
    uint32_t size() const;
    bool empty() const;
    const LRUListItem& operator[](uint32_t index) const;

    uint32_t findValue(const string_t& value) const;

    // manipulation
    void pushNew(LRUListItem* item);
    void pushNewLimit(LRUListItem* item);
    void moveToFront(uint32_t index);
    void clear();

    static bool isSameValue(const string_t& a, const string_t& b);

 private:
    ptr_vector<LRUListItem> items;
};

extern LRUList predef_shiplabels, predef_planetlabels, predef_search;
extern LRUList lru_shiplabels, lru_planetlabels, lru_search;

extern const struct ModuleInitializer mod_lru[];

/**************************** Inline Functions ***************************/

/** Create empty list. */
inline LRUList::LRUList()
{ }

/** Destructor. */
inline LRUList::~LRUList()
{ }

/** Append new item.
    \param item Newly-allocated item. Will be owned by the list. */
inline void
LRUList::pushNew(LRUListItem* item)
{
    items.push_back(item);
}

/** Clear this list. */
inline void
LRUList::clear()
{
    items.clear();
}



/** Create item.
    \param name Name
    \param value Value */
inline
LRUListItem::LRUListItem(string_t name, string_t value)
    : name(name),
      value(value)
{ }

/** Get name of this item. */
inline const string_t&
LRUListItem::getName() const
{
    return name;
}

/** Get value of this item. */
inline const string_t&
LRUListItem::getValue() const
{
    return value;
}

/** Get number of elements in this list. */
inline uint32_t
LRUList::size() const
{
    return items.size();
}

/** Check for emptiness. */
inline bool
LRUList::empty() const
{
    return items.empty();
}

/** Get item by index.
    \param index [0,size()) Index */
inline const LRUListItem&
LRUList::operator[](uint32_t index) const
{
    return *items[index];
}

#endif
