/**
  *  \file flak/vis-iso.cc
  *  \brief ISO-3D visualizer
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <climits>
#include "flak/vis-sprites.h"
#include "flak/vis.h"
#include "game/formula.h"
#include "game/spec.h"
#include "gfx/blit.h"
#include "gfx/palette.h"
#include "gfx/sprite3d.h"
#include "resmgr/resid.h"
#include "resmgr/resmgr.h"

namespace {
    enum {
        SHIP_RADIUS = 5
    };

    //! Isometric Ship Sprite.
    class FlakIsoShipSprite : public GfxSprite3D {
        FlakIsoVisualizer& vis;
        FlakBattle::ship_t ship;
        int angle;
        int timer;
        Ptr<GfxPixmap> pix;
     public:
        FlakIsoShipSprite(FlakIsoVisualizer& vis,
                          FlakBattle::ship_t ship)
            : GfxSprite3D(FlakIsoVisualizer::layer_Ships),
              vis(vis), ship(ship), angle(-1), timer(0)
            { tick(); }
        void draw(GfxCanvas& can);
        void tick();
    };
}

static const int num_angles     = 16;
static const int min_turn_delay = 5;

/** Get rotated ship picture.
    \param hull hull number
    \param angle angle, [0, num_angles). */
static Ptr<GfxPixmap>
getShipPicture(int hull, int angle)
{
    ResManager& mgr = ResManager::getInstance();
    const ResId id(res::ship_rotated, hull, angle+1);
    Ptr<GfxPixmap> pix = mgr.getPixmap(id);
    if (pix)
        return pix;
    pix = mgr.getPixmap(ResId(res::rship, ::getHull(hull).getInternalPictureNumber()));
    if (pix)
        pix = pix->convertToScreenFormat();

    for (int i = 0; i < num_angles; ++i) {
        if (pix == 0) {
            mgr.addPixmap(ResId(res::ship_rotated, hull, i+1), 0);
        } else {
            double rangle = i * (2 * PI / num_angles);
            const int wi = 32, he = 32;
            int x1 = int (wi * std::cos(rangle));
            int y1 = int (wi * std::sin(rangle));
            int x2 = int (he * std::cos(rangle - PI/2));
            int y2 = int (he * std::sin(rangle - PI/2));
            GfxRect new_size = computeStretchRotateBBox(x1, y1, x2, y2);

            Ptr<GfxPixmap> result = new GfxPixmap(new_size.w, new_size.h, *pix);
            doStretchRotateBlit(*pix,
                                *result,
                                GfxRect(*pix),
                                GfxRect(*result),
                                -new_size.x,
                                -new_size.y,
                                x1, y1, x2, y2);
            result->setColorKey(result->getPixel(GfxPoint(0, 0)));

            mgr.addPixmap(ResId(res::ship_rotated, hull, i+1),
                          result->cloneCropped(result->getColorKey()));
        }
    }
    return mgr.getPixmap(id);
}

void
FlakIsoShipSprite::draw(GfxCanvas& can)
{
    if (vis.getBattle().getShipByNumber(ship).isAlive()) {
        if (pix) {
            pix->blit(can, getPos().x, getPos().y);
        } else {
            const bool planet = vis.getBattle().getShipByNumber(ship).isPlanet();
            GfxContext ctx;
            ctx.useCanvas(can);
            ctx.useColorScheme(GfxStandardColorScheme::instance);
            ctx.setColor(!planet ? COLOR_BLACK : COLOR_BLUE);
            drawFilledCircle(ctx, getPos().getCenter(), SHIP_RADIUS);
            ctx.setColor(!planet ? COLOR_WHITE : COLOR_GREEN);
            drawCircle(ctx, getPos().getCenter(), SHIP_RADIUS);
        }
    }
}

void
FlakIsoShipSprite::tick()
{
    const FlakShip&     sh = vis.getBattle().getShipByNumber(ship);
    const FlakScreenPos sp = vis.getScreenPosition(sh.getPos());

    if (!sh.isAlive())
        return;

    if (!sh.isPlanet()) {
        /* Figure out movement target. If ship has an enemy, let it
           point there. Otherwise, let it point to the origin.
           Originally, ships moved there when they had no enemy. */
        FlakScreenPos ep;
        if (FlakShip* enemy_ptr = sh.fleet_link->status.enemy_ptr)
            ep = vis.getScreenPosition(enemy_ptr->getPos());
        else
            ep = vis.getScreenPosition(FlakPos(0, 0, sh.getPos().z));

        /* Compute angle. If target and ship are on same screen
           position, pick an arbitrary angle. The previous behaviour
           was to simply do nothing in this case, which causes the
           ship sprite to be replaced by the fallback circle when
           people use rewind. */
        int new_angle;
        if (ep.x != sp.x || ep.y != sp.y)
            new_angle = int(std::atan2(double(ep.y - sp.y), double(ep.x - sp.x)) * num_angles / (2*PI) + num_angles + 0.5) % num_angles;
        else
            new_angle = 0;

        /* Turn the ship. We use a timer to limit the amount of turn
           steps per tick. */
        int angle_diff = (num_angles + new_angle - angle) % num_angles;
        int old_angle = angle;
        if (angle == -1) {
            angle = new_angle;
            timer = 0;
        } else if (angle_diff == 0) {
            timer = 0;
        } else if (timer < min_turn_delay) {
            if (vis.isTimeAdvance())
                timer++;
        } else {
            if (angle_diff < num_angles/2) {
                ++angle;
                if (angle >= num_angles)
                    angle = 0;
            } else {
                --angle;
                if (angle < 0)
                    angle = num_angles-1;
            }
            timer = 0;
        }
        if (angle != old_angle) {
            pix = getShipPicture(sh.data.hull, angle);
            update();
        }
        if (pix) {
            setPos(GfxRect(sp.x - pix->getWidth()/2, sp.y - pix->getHeight()/2,
                           pix->getWidth(), pix->getHeight()),
                   sp.z);
            return;
        }
    }

    setPos(GfxRect(sp.x - SHIP_RADIUS, sp.y - SHIP_RADIUS, 2*SHIP_RADIUS+1, 2*SHIP_RADIUS+1), sp.z);
}

/*************************** FlakIsoVisualizer ***************************/

/*! \class FlakIsoVisualizer
    This displays the fight in a simple isometric 3-D view. */

/** Constructor.
    \param battle the fight to display
    \param area screen area to use */
FlakIsoVisualizer::FlakIsoVisualizer(FlakBattle& battle, GfxRect area)
    : FlakFancyVisualizer(battle, area),
      center_x(0), center_y(0), range(0), ship_sprites(battle.getNumShips())
{
    computeViewport();

    Ptr<GfxPixmap> pix = requestPixmap(ResId("flak.background"));
    if (pix)
        getManager().addSprite(new GfxPixmapSprite3D(0, 0, INT_MAX, layer_Scene, pix));
}

FlakIsoVisualizer::~FlakIsoVisualizer()
{ }

void
FlakIsoVisualizer::computeViewport()
{
    center_x = center_y = 0;

    int32 min_range = 19000;
    for (fleet_t i = 0; i < the_battle.getNumFleets(); ++i) {
        FlakFleet& f = the_battle.getFleetByNumber(i);
        min_range = std::max(min_range,
                             int32(std::max(std::abs(f.status.position.x - center_x),
                                            std::abs(f.status.position.y - center_y))));
    }
    range = 4*min_range + 2000;
}

void
FlakIsoVisualizer::everyTickAction()
{ }

void
FlakIsoVisualizer::createSprites(bool is_startup)
{
    if (is_startup)
        computeViewport();

    // create ships
    for (ship_t i = 0, max = getBattle().getNumShips(); i < max; ++i) {
        ship_sprites[i] = new FlakIsoShipSprite(*this, i);
        manager.addSprite(ship_sprites[i]);
    }

    // create objects
    for (int player = 1; player <= FLAK_NUM_OWNERS; ++player)
        for (FlakObject* o = getBattle().getPlayerObjects(player); o != 0; o = o->next)
            createObject(*o, FlakBattle::no_ship);
}

FlakScreenPos
FlakIsoVisualizer::getScreenPosition(FlakPos fp)
{
    const GfxRect& a = getManager().getBBox();
    return FlakScreenPos(a.getCenter().x + (a.w * (fp.x - center_x - fp.y + center_y) / range),
                         a.getCenter().y - (a.h * (fp.x - center_x + fp.y - center_y) / (3*range)) - fp.z*25000/range,
                         range + fp.x/2 - fp.y/2);
}

Ptr<GfxSprite3D>
FlakIsoVisualizer::getShipSprite(ship_t ship)
{
    return ship_sprites[ship];
}

FlakFancyVisualizer::ship_t
FlakIsoVisualizer::getShipAtPosition(GfxPoint pt)
{
    ship_t found = FlakBattle::no_ship;
    int found_z = 0;
    for (ship_t i = 0; i < ship_sprites.size(); ++i) {
        if (getBattle().getShipByNumber(i).isAlive()
            && ship_sprites[i]
            && ship_sprites[i]->getPos().contains(pt))
        {
            int this_z = ship_sprites[i]->getPosZ();
            if (found == FlakBattle::no_ship || this_z < found_z) {
                found = i;
                found_z = this_z;
            }
        }
    }
    return found;
}

FlakFancyVisualizer::EventResult
FlakIsoVisualizer::handleEvent(const UIEvent& evt)
{
    (void) evt;
    return eNotHandled;
}

void
FlakIsoVisualizer::createObject(FlakObject& obj, ship_t source)
{
    if (source != FlakBattle::no_ship)
        markShipChanged(source);
    setObjectSprite(obj, FlakSimpleObjectSprite::create(*this, obj), true);
}
