/**
  *  \file int/bytecode.h
  *  \brief Interpreter: Bytecode Object
  *
  *  \todo Can we make referred BCOs literals? pro: this would save a special
  *  case and an addressing mode. con: we would need to be able to serialize
  *  BCOs in IntDataSegment.
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_BYTECODE_H
#define PCC_V2_INT_BYTECODE_H

#include <vector>
#include <algorithm>
#include <functional>
#include "cpluslib/types.h"
#include "int/data.h"
#include "int/opcode.h"
#include "int/names.h"
#include "cpluslib/smartptr.h"

class IntBytecodeObject;
struct IntCompilationContext;

class IntVMSaveContext;
class IntVMLoadContext;

/** Reference to an IntBytecodeObject. */
typedef Ptr<IntBytecodeObject> IntBCORef;

/** Bytecode object (BCO). Bytecode objects contain code for execution.
    It contains the following elements:
    - actual code
    - a list of literals referred to by the code (only small integer literals
      can be encoded directly in the code, others are loaded from this table).
    - a list of names (symbols) referred to by the code.
    - a list of other BCOs referred to by the code. This means only subroutines
      <em>defined</em> by this BCO, not <em>called</em> subroutines.
    - a list of predeclared identifiers
    - additional information about this BCO as a subroutine ("is procedure"
      flag, argument counts).
    - optional additional information about this BCO's source code (aka debug
      information), i.e. file name and line/address associations. We assume
      that each BCO is compiled from a single file, that is, we don't have an
      "#include"-style preprocessor, but a "Load" instruction executed at
      runtime.

    For the benefit of code generation, the BCO can also contain symbolic
    labels.

    BCOs use reference counting. Each IntExecutionFrame executing a BCO refers
    to it, as well as the symbol table entry holding it, as well as another
    BCO which happened to define it. Circular references are impossible. For
    example, assume we are executing "foo" and are at the line containing "bar":
    <pre>
      Sub foo
        Sub bar
          Print "hi"
        EndSub
        bar
      EndSub
    </pre>
    The BCO "foo" will be referenced from the symbol table and the executing
    frame. The BCO "bar" will be referenced from the symbol table (it has just
    been defined), from the executing frame, and from the "foo" BCO.

    BCOs contain a list of predeclared identifiers (see getLocalNames). These
    are the names of the parameters used to invoke the parameters, plus the
    names of predeclared local variables. Predeclaring a variable will give it
    a known address at compile time, allowing faster code to be generated.
    The compiler will use this instead of a <code>dimloc</code> instruction
    when it can prove that it's safe to do so. */
class IntBytecodeObject {
    friend class IntVMSaveContext;
    friend class IntVMLoadContext;
 public:
    typedef std::vector<IntOpcode>::size_type pc_t;
    typedef uint16_t label_t;

    IntBytecodeObject();
    ~IntBytecodeObject();

    void     addArgument(string_t name, bool optional);
    void     addLocalVariable(const string_t& name);
    bool     hasLocalVariable(const string_t& name);
    void     setIsProcedure(bool flag);
    void     setIsVarargs(bool flag);
    int      getMinArgs() const;
    int      getMaxArgs() const;
    bool     isProcedure() const;
    bool     isVarargs() const;
    string_t getName() const;
    void     setName(string_t name);
    string_t getFileName() const;
    void     setFileName(string_t file_name);
    void     addLineNumber(uint32_t line);
    uint32_t getLineNumber(pc_t pc) const;

    label_t  makeLabel();
    void     addInstruction(IntOpcode::Major major, uint8_t minor, uint16_t arg);
    void     addVerifyInstruction(IntOpcode::Major major, uint8_t minor, uint32_t arg);
    void     addVariableReferenceInstruction(IntOpcode::Major major, const string_t& name, const IntCompilationContext& cc);
    void     addLabel(label_t label);
    void     addJump(uint8_t flags, label_t label);
    void     addPushLiteral(IntValue* literal);
    uint16_t addName(string_t name);
    bool     hasName(string_t name) const;
    bool     hasUserCall() const;
    void     relocate();
    void     compact();
    void     copyLocals(const IntBytecodeObject& other);
    void     append(IntBytecodeObject& other);

    pc_t     getNumInstructions() const;
    uint32_t getNumLabels() const;
    pc_t     getJumpTarget(uint8_t minor, uint16_t arg) const;
    IntOpcode&       operator()(pc_t index);
    const IntOpcode& operator()(pc_t index) const;
    string_t getDisassembly(pc_t index) const;

    IntValue* getLiteral(uint16_t index) const;
    const string_t& getName(uint16_t index) const;
    const IntVariableNames& getLocalNames() const;

 private:
    IntDataSegment         data;   ///< Literals referenced in bytecode.
    IntVariableNames       names;  ///< Names referenced in bytecode.
    std::vector<IntOpcode> code;   ///< Actual code.
    label_t                num_labels;
    IntVariableNames       local_names;
    int                    min_args, max_args;
    bool                   is_procedure;
    bool                   is_varargs;
    string_t               name;
    string_t               file_name;
    std::vector<uint32_t>  line_numbers; ///< Line numbers. Pairs of address,line.
};

/**************************** Inline Functions ***************************/

namespace std {
    /** Total order for IntBCORef. This allows to make a std::map
        using IntBCORef's as key. We don't need any particular order,
        so we just use the address order. */
    template<>
    struct less<IntBCORef> {
        bool operator()(IntBCORef a, IntBCORef b) const
            { return less<void*>()(a.ptr(), b.ptr()); }
    };
}

/** Add local variable. addLocalVariable() should not be followed by addArgument(). */
inline void
IntBytecodeObject::addLocalVariable(const string_t& name)
{
    local_names.add(name);
}

/** Check for local variable.
    \param name Name to check */
inline bool
IntBytecodeObject::hasLocalVariable(const string_t& name)
{
    return local_names.lookup(name) != local_names.nil;
}

/** Set the "is procedure" flag. If set, the generated code is a procedure without
    result, which leaves the stack as-is. If clear, the generated code is a function
    which generates a single result on the stack. */
inline void
IntBytecodeObject::setIsProcedure(bool flag)
{
    is_procedure = flag;
}

/** Set the "is varargs" flag. If set, arguments exceeding the maximum number of
    args are wrapped into an array. */
inline void
IntBytecodeObject::setIsVarargs(bool flag)
{
    is_varargs = flag;
}

/** Get minimum number of expected arguments. */
inline int
IntBytecodeObject::getMinArgs() const
{
    return min_args;
}

/** Get maximum number of expected arguments. */
inline int
IntBytecodeObject::getMaxArgs() const
{
    return max_args;
}

/** Get the "is procedure" flag. \see setIsProcedure */
inline bool
IntBytecodeObject::isProcedure() const
{
    return is_procedure;
}

/** Get the "is varargs" flag. \see setIsVarargs */
inline bool
IntBytecodeObject::isVarargs() const
{
    return is_varargs;
}

/** Check whether we already know the specified name. */
inline bool
IntBytecodeObject::hasName(string_t name) const
{
    return names.lookup(name) != names.nil;
}

/** Get number of instructions in this bytecode object. */
inline IntBytecodeObject::pc_t
IntBytecodeObject::getNumInstructions() const
{
    return code.size();
}

/** Get number of symbolic labels in this bytecode object. */
inline uint32_t
IntBytecodeObject::getNumLabels() const
{
    return num_labels;
}

/** Read/write access to instruction by address. */
inline IntOpcode&
IntBytecodeObject::operator()(pc_t index)
{
    return code[index];
}

/** Read access to instruction by address. */
inline const IntOpcode&
IntBytecodeObject::operator()(pc_t index) const
{
    return code[index];
}

/** Get local name list. These are the parameter names and the
    predeclared local variables. */
inline const IntVariableNames&
IntBytecodeObject::getLocalNames() const
{
    return local_names;
}

/** Get literal from literal table. */
inline IntValue*
IntBytecodeObject::getLiteral(uint16_t index) const
{
    /* No range check required; it is done by IntDataSegment. */
    return data[index];
}

/** Get name from name table. */
inline const string_t&
IntBytecodeObject::getName(uint16_t index) const
{
    return names.lookupAddress(index);
}

#endif
