#!/usr/bin/perl -w
#
#  PCC2 Object File Stripper
#
#  This removes debug information (=line numbers) from PCC2 object files.
#

use strict;

foreach (@ARGV) {
    if (/^-/) {
        die "No options accepted.";
    } else {
        strip($_);
    }
}

sub strip {
    my $fn = shift;
    my $result;
    open FILE, "< $fn" or die "$fn: $!\n";
    binmode FILE;

    # Copy header
    my $header;
    if (read(FILE, $header, 14) != 14)          { die "File format error" }
    if (substr($header, 0, 6) ne "CCobj\032")   { die "File format error" }
    if (unpack("v",substr($header, 8, 2)) != 4) { die "File format error" }
    $result .= $header;

    # Translate objects
    while (my $p = loadObject(\*FILE)) {
        # Remove debug information
        if ($p->{type} == 1) {
            $p->{properties}[7] = [0,'']
              if exists $p->{properties}[7];
            $p->{properties}[8] = [0,'']
              if exists $p->{properties}[8];
        }

        # Remove empty trailing properties
        while ($p->{properties}[-1][0] == 0 && $p->{properties}[-1][1] eq '') {
            pop @{$p->{properties}};
        }

        $result .= objToString($p);
    }

    # Write it out
    printf "%-30s  %d -> %d bytes\n", $fn, tell(FILE), length($result);
    close FILE;
    open FILE, "> $fn" or die "$fn: $!\n";
    binmode FILE;
    print FILE $result;
    close FILE;
}

sub loadObject {
    my $fd = shift;

    # Read header
    my $header;
    my $nbytes = read($fd, $header, 16);
    if ($nbytes == 0) { return undef; }
    if ($nbytes != 16) { die "File format error" }

    # Read content
    my ($typ, $id, $size, $nprop) = unpack "V*", $header;
    if ($size < $nprop*8) { die "File format error" }
    # print "Object: type=$typ, id=$id, size=$size, nprop=$nprop\n";

    # Read property headers
    my $phead;
    my @phead;
    if (read($fd, $phead, 8*$nprop) != 8*$nprop) { die "File format error" }
    @phead = unpack "V*", $phead;
    $size -= 8*$nprop;

    # Read properties
    my $props = [];
    for (my $i = 0; $i < $nprop; ++$i) {
        my $count = shift @phead;
        my $psize = shift @phead;
        # print "  Property $i: $count elems, $psize bytes\n";
        if ($psize > $size) { die "File format error" }
        $size -= $psize;
        my $pdata;
        if (read($fd, $pdata, $psize) != $psize) { die "File format error" }
        push @$props, [$count,$pdata];
    }

    if ($size != 0) { die "File format error" }

    return { id => $id,
             type => $typ,
             properties => $props };
}

sub objToString {
    my $p = shift;

    # Convert properties
    my $pdata = "";
    my $phead = "";
    foreach (@{$p->{properties}}) {
        $pdata .= $_->[1];
        $phead .= pack("VV", $_->[0], length($_->[1]));
    }

    # Generate output
    return pack("VVVV", $p->{type}, $p->{id}, length($pdata)+length($phead), scalar(@{$p->{properties}}))
      . $phead
        . $pdata;
}
